%% INTRODUCTION
% This script enables a simple test case of prediction with quantified
% uncertainty. Given timing measurements of a tennis ball dropping from 
% known heights and model for time as a function of height, can we predict 
% the time it will take a tennis ball dropped from the top of Hoover Tower 
% to hit the ground?
clear all; close all; clc;

%% MEASUREMENTS
% Collect the timing measurements. (THIS PART IS HARD!)
ndata = 3;
heights = [2.7432; 4.7244; 7.4676];
times = [0.6 0.8 0.6; 0.9 1.0 0.9; 1.1 1.2 1.3];

%% MEASUREMENT ERROR MODEL
% Assume the random error model for the timing measurement is normally 
% distributed with zero mean and a standard deviation.
nsamples = 5000;
mu = 0; 
sigma = 0.1;
error_samples = mu + sigma*randn(ndata,nsamples);

%% COMPUTATIONAL MODEL - THE FIRST ATTEMPT
% Assume the model for time as a function of height can be expressed by the
% linear relationship t = x1 + x2*h. 
time_model_1 = @(x1,x2,height) x1 + x2.*height;

%% VERIFICATION
% Verify the computational code with an exact solution.
if time_model_1(1,1,1) == 2
    [y,Fs]=wavread('verified.wav'); sound(y,Fs); clear y Fs
end

%% PARAMETER ESTIMATION / MODEL CALIBRATION
% We use a least squares method to infer the parameters x1 and x1.
A = [ones(ndata,1) heights];
b = mean(times,2);

% Draw samples of the model parameters x1 and x2 that are consistent with 
% the measured heights. 
param_samples = zeros(2,nsamples);
for i = 1:nsamples
    e = error_samples(:,i);
    t = b + e;
    param_samples(:,i) = (A'*A)\(A'*t);
end

% Plot marginal densities for the parameters.
[m1,pout1]=hist(param_samples(1,:),30);
[m2,pout2]=hist(param_samples(2,:),30);
figure;
subplot(1,2,1);
bar(pout1,m1/sum(m1));
set(gca,'FontSize',16);
xlabel('x1');
subplot(1,2,2);
bar(pout2,m2/sum(m2));
set(gca,'FontSize',16);
xlabel('x2');

%% FORWARD PROPAGATION
% Sample the model at the sampled parameter values.
model_samples = zeros(ndata,nsamples);
for i = 1:nsamples
    x = param_samples(:,i);
    for j = 1:ndata
        model_samples(j,i) = time_model_1(x(1),x(2),heights(j));
    end 
end

% Plot the model predictions with the measurements.
figure;
plot(heights,model_samples,'rx',...
	heights,times,'gs',...
    'MarkerFaceColor','g','MarkerSize',10);
set(gca,'FontSize',16);
xlabel('Height (m)');
ylabel('Time (s)');
title('Model 1');

%% PREDICTION WITH QUANTIFIED UNCERTAINTY
% Predict the time it would take to fall from Hoover Tower.
hoover_height = 86.86;
hsamples = zeros(nsamples,1);
for i = 1:nsamples
    x = param_samples(:,i);
    hsamples(i) = time_model_1(x(1),x(2),hoover_height);
end

% Display a histogram of the predictions.
[n1,xout1]=hist(hsamples,30);
figure;
bar(xout1,n1/sum(n1),'r');
set(gca,'FontSize',16);
xlabel('Time (s)');
title('Model 1');

%% COMPUTATIONAL MODEL - THE SECOND ATTEMPT
% Suppose now a physicist gives us a better model, t = x1*sqrt(h).
time_model_2 = @(x1,height) x1.*sqrt(height);

%% VERIFICATION
% Verify computational code with an exact solution.
if time_model_2(1,4) == 2
    [y,Fs]=wavread('verified.wav'); sound(y,Fs); clear y Fs
end

%% PARAMETER ESTIMATION / MODEL CALIBRATION
% Get samples of the model parameters using a similar least squares method.
param_samples = zeros(nsamples,1);
for i = 1:nsamples
    e = error_samples(:,i);
    t = b + e;
    param_samples(i) = (sqrt(heights)'*t)/(sqrt(heights)'*sqrt(heights));
end

% Plot the density of the parameter.
[m,pout]=hist(param_samples,30);
figure;
bar(pout,m/sum(m));
set(gca,'FontSize',16);
xlabel('x1');
title('Model 2 Parameter');

%% FORWARD PROPAGATION
% Sample the new model at the sampled parameter values.
for i = 1:nsamples
    x = param_samples(i);
    for j = 1:ndata
        model_samples(j,i) = time_model_2(x,heights(j));
    end 
end

% Plot the new model predictions with the measurements.
figure;
plot(heights,model_samples,'rx',...
	heights,times,'gs',...
    'MarkerFaceColor','g','MarkerSize',10);
set(gca,'FontSize',16);
xlabel('Height (m)');
ylabel('Time (s)');
title('Model 2');

%% PREDICTION WITH QUANTIFIED UNCERTAINTY
% Predict the time it would take to fall from Hoover Tower with the new and
% improved model.
hsamples = zeros(nsamples,1);
for i = 1:nsamples
    x = param_samples(i);
    hsamples(i) = time_model_2(x,hoover_height);
end

% Display a histogram of the predictions with the new model.
[n2,xout2]=hist(hsamples,30);
figure;
bar(xout2,n2/sum(n2),'r');
set(gca,'FontSize',16);
xlabel('Time (s)');
title('Model 2');

%% COMPARISON
% Compare the two model predictions.
figure;
hold on
bar(xout2,n2/sum(n2),'r');
bar(xout1,n1/sum(n1),'b');
set(gca,'FontSize',16);
xlabel('Time (s)');
title('Comparison');
legend('Model 2','Model 1');
hold off

